package dev.fitko.fitconnect.client.factory;

import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.exceptions.InitializationException;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.steps.sender.StartStep;
import dev.fitko.fitconnect.client.steps.subscriber.RequestSubmissionStep;
import dev.fitko.fitconnect.core.SubmissionSender;
import dev.fitko.fitconnect.core.SubmissionSubscriber;
import dev.fitko.fitconnect.core.auth.DefaultOAuthService;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.destination.DestinationRestService;
import dev.fitko.fitconnect.core.events.EventLogApiService;
import dev.fitko.fitconnect.core.events.SecurityEventTokenService;
import dev.fitko.fitconnect.core.http.ProxyConfig;
import dev.fitko.fitconnect.core.submission.SubmissionApiService;
import dev.fitko.fitconnect.core.validation.DefaultValidationService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.client.RestTemplate;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.text.ParseException;
import java.util.Objects;

/**
 * Factory that constructs clients for {@link Sender} and {@link Subscriber}.
 */
public final class ClientFactory {

    private static final Logger LOGGER = LoggerFactory.getLogger(ClientFactory.class);

    public static final String CONFIG_ENV_KEY_NAME = "FIT_CONNECT_CONFIG";
    public static final String METADATA_SCHEMA_JSON = "/metadata_schema.json";
    public static final String SET_EVENT_SCHEMA_JSON = "/set_payload.schema.json";

    private ClientFactory() {
    }

    /**
     * Create a new {@link SenderClient} that is automatically configured via the config.yml file.
     *
     * @return the sender client
     */
    public static StartStep senderClient() {
        final ApplicationConfig config = loadConfig();
        return senderClient(config);
    }

    /**
     * Create a new {@link SenderClient} that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the sender client
     */
    public static StartStep senderClient(final ApplicationConfig config) {
        LOGGER.info("Initializing sender client ...");
        final Sender sender = getSender(config);
        return SenderClient.build(sender, config.getCurrentEnvironment());
    }


    /**
     * Create a new {@link SubscriberClient} that is automatically configured via config.yml file.
     *
     * @return the subscriber client
     */
    public static RequestSubmissionStep subscriberClient() {
        final ApplicationConfig config = loadConfig();
        return subscriberClient(config);
    }

    /**
     * Create a new {@link SubscriberClient} that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the subscriber client
     */
    public static RequestSubmissionStep subscriberClient(final ApplicationConfig config) {
        LOGGER.info("Initializing subscriber client ...");
        final Subscriber subscriber = getSubscriber(config);
        final SubscriberConfig subscriberConfig = config.getSubscriberConfig();

        LOGGER.info("Reading private key from {} ", subscriberConfig.getPrivateDecryptionKeyPath());
        final String privateKeyPath = readPath(subscriberConfig.getPrivateDecryptionKeyPath(), "Decryption Key");
        final RSAKey privateKey = readRSAKeyFromString(privateKeyPath);

        return SubscriberClient.builder(subscriber, privateKey);
    }

    private static Subscriber getSubscriber(final ApplicationConfig config) {
        final CryptoService cryptoService = getCryptoService();
        final ValidationService validator = getValidationService(config);
        final RestTemplate restTemplate = getRestTemplate(config);
        final OAuthService authService = getSubscriberConfiguredAuthService(config, restTemplate);
        final EventLogService eventLogService = getEventLogService(config, restTemplate, authService);
        final SubmissionService submissionService = getSubmissionService(config, restTemplate, authService);
        final SecurityEventService setService = getSecurityEventTokenService(config, validator);
        return new SubmissionSubscriber(submissionService, eventLogService, cryptoService, validator, setService);
    }

    private static Sender getSender(final ApplicationConfig config) {
        final CryptoService cryptoService = getCryptoService();
        final RestTemplate restTemplate = getRestTemplate(config);
        final ValidationService validator = getValidationService(config);
        final OAuthService authService = getSenderConfiguredAuthService(config, restTemplate);
        final EventLogService eventLogService = getEventLogService(config, restTemplate, authService);
        final SubmissionService submissionService = getSubmissionService(config, restTemplate, authService);
        final DestinationService destinationService = getDestinationApiService(config, restTemplate, authService);
        return new SubmissionSender(destinationService, submissionService, eventLogService, cryptoService, validator);
    }

    private static OAuthService getSenderConfiguredAuthService(final ApplicationConfig config, final RestTemplate restTemplate) {
        final String clientId = config.getSenderConfig().getClientId();
        final String clientSecret = config.getSenderConfig().getClientSecret();
        return new DefaultOAuthService(restTemplate, clientId, clientSecret, config.getOAuthTokenEndpoint());
    }

    private static OAuthService getSubscriberConfiguredAuthService(final ApplicationConfig config, final RestTemplate restTemplate) {
        final String clientId = config.getSubscriberConfig().getClientId();
        final String clientSecret = config.getSubscriberConfig().getClientSecret();
        return new DefaultOAuthService(restTemplate, clientId, clientSecret, config.getOAuthTokenEndpoint());
    }

    private static SubmissionService getSubmissionService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new SubmissionApiService(authService, restTemplate, config);
    }

    private static EventLogService getEventLogService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new EventLogApiService(authService, restTemplate, config);
    }

    private static DestinationRestService getDestinationApiService(final ApplicationConfig config, final RestTemplate restTemplate, final OAuthService authService) {
        return new DestinationRestService(authService, restTemplate, config);
    }

    private static ValidationService getValidationService(final ApplicationConfig config) {
        final MessageDigestService messageDigestService = getMessageDigestService();
        LOGGER.info("Reading metadata schema from {} ", METADATA_SCHEMA_JSON);
        final String metadataSchema = readSchema(METADATA_SCHEMA_JSON);
        LOGGER.info("Reading set payload schema from {} ", SET_EVENT_SCHEMA_JSON);
        final String setPayloadSchema = readSchema(SET_EVENT_SCHEMA_JSON);
        return new DefaultValidationService(messageDigestService, metadataSchema, setPayloadSchema, config);
    }

    private static CryptoService getCryptoService() {
        final MessageDigestService messageDigestService = getMessageDigestService();
        return new JWECryptoService(messageDigestService);
    }

    private static RestTemplate getRestTemplate(final ApplicationConfig config) {
        final ProxyConfig proxyConfig = new ProxyConfig(config.getHttpProxyHost(), config.getHttpProxyPort());
        return proxyConfig.proxyRestTemplate();
    }

    private static MessageDigestService getMessageDigestService() {
        return new HashService();
    }

    private static SecurityEventTokenService getSecurityEventTokenService(final ApplicationConfig config, final ValidationService validationService) {
        LOGGER.info("Reading private signing key from {} ", config.getSubscriberConfig().getPrivateSigningKeyPath());
        final String signingKey = readPath(config.getSubscriberConfig().getPrivateSigningKeyPath(), "Signing Key");
        final RSAKey rsaKey = readRSAKeyFromString(signingKey);
        return new SecurityEventTokenService(config, validationService, rsaKey);
    }

    private static RSAKey readRSAKeyFromString(final String key) {
        try {
            return RSAKey.parse(JWK.parse(key).toJSONObject());
        } catch (final ParseException | NullPointerException e) {
            throw new InvalidKeyException("Key could not be parsed", e);
        }
    }

    private static String readPath(final String pathName, final String errorKey) {
        try {
            return Files.readString(Path.of(pathName));
        } catch (final IOException e) {
            LOGGER.error("Could not read {} from path '{}'. Please provide a valid path in your config.yml", errorKey, pathName);
            throw new InitializationException(e.getMessage(), e);
        }
    }

    private static String readSchema(final String schema) {
        try (InputStream inputStream = ClientFactory.class.getResourceAsStream(schema)) {
            return new String(Objects.requireNonNull(inputStream).readAllBytes());
        } catch (final IOException e) {
            LOGGER.error("Schema '{}' could not be loaded", schema);
            throw new InitializationException(e.getMessage(), e);
        }
    }

    private static Path readConfigFromEnvironment() {
        try {
            return Path.of(System.getenv(CONFIG_ENV_KEY_NAME));
        } catch (final NullPointerException e) {
            LOGGER.error("Environment variable {} could not be loaded", CONFIG_ENV_KEY_NAME);
            throw new InitializationException(e.getMessage(), e);
        }
    }

    private static ApplicationConfig loadConfig() {
        try {
            final Path configPath = readConfigFromEnvironment();
            final ApplicationConfig applicationConfig = ApplicationConfigLoader.loadConfig(configPath);
            LOGGER.info("Using sdk environment config `{}` ", applicationConfig.getActiveEnvironment().getName());
            return applicationConfig;
        } catch (final InitializationException e) {
            LOGGER.error("Config could not be loaded, please provide a 'config.yml' by setting the environment variable 'FIT_CONNECT_CONFIG'");
            throw new InitializationException(e.getMessage(), e);
        }
    }
}
